# coding: utf-8

# Copyright 2019 Matthieu Dien and Martin Pépin
# Distributed under the license GNU GPL v3 or later
# See LICENSE.txt for more informations

r"""

Index of examples

.. autosummary::

   expnums
   knr19
   mondrian

"""
